"use strict";
const utils_1 = require("../metadata/utils");
const decorators_1 = require('../constants/decorators');
const decorator_type_1 = require('../enums/decorator-type');
const principalContext_1 = require('../../security/auth/principalContext');
const Utils = require("../utils");
const model_entity_1 = require('../dynamic/model-entity');
const constantKeys_1 = require('../constants/constantKeys');
var Q = require('q');
function transaction(params) {
    params = params || {};
    return function (target, propertyKey, descriptor) {
        utils_1.MetaUtils.addMetaData(target, {
            decorator: decorators_1.Decorators.TRANSACTION,
            decoratorType: decorator_type_1.DecoratorType.METHOD,
            params: params,
            propertyKey: propertyKey
        });
        var originalMethod = descriptor.value;
        descriptor.value = function () {
            let transaction = principalContext_1.PrincipalContext.get(constantKeys_1.ConstantKeys.transaction);
            let args = [];
            args = Array.apply(null, arguments);
            if (!transaction && this.path) {
                let entityService = Utils.entityService(model_entity_1.pathRepoMap[this.path].modelType);
                return entityService.startTransaction().then(transaction => {
                    principalContext_1.PrincipalContext.save(constantKeys_1.ConstantKeys.transaction, transaction);
                    return executeOriginalMethod.apply(this, [originalMethod, args]).then(result => {
                        return entityService.commitTransaction(principalContext_1.PrincipalContext.get(constantKeys_1.ConstantKeys.transaction)).then(commit => {
                            return result;
                        });
                    });
                }).catch(exc => {
                    return entityService.rollbackTransaction(principalContext_1.PrincipalContext.get(constantKeys_1.ConstantKeys.transaction)).then(result => {
                        principalContext_1.PrincipalContext.save(constantKeys_1.ConstantKeys.transaction, null);
                        throw exc;
                    });
                });
            }
            else {
                return executeOriginalMethod.apply(this, [originalMethod, args]);
            }
        };
        return descriptor;
    };
}
exports.transaction = transaction;
function executeOriginalMethod(originalMethod, args) {
    let ret = originalMethod.apply(this, args);
    if (Utils.isPromise(ret)) {
        return ret;
    }
    else {
        return Q.when(ret);
    }
}

//# sourceMappingURL=transaction.js.map
