"use strict";
const Enumerable = require('linq');
const utils_1 = require("../../core/metadata/utils");
const decorators_1 = require('../../core/constants/decorators');
const model_entity_1 = require('../dynamic/model-entity');
const Utils = require("../utils");
exports.getQueryOptionsFromQuery = (repository, query) => {
    let primaryKey;
    var metaDataMap = utils_1.MetaUtils.getMetaData(repository.entity.model, decorators_1.Decorators.COLUMN);
    let entityService = Utils.entityService(model_entity_1.pathRepoMap[repository.path].modelType);
    for (var field in metaDataMap) {
        var fieldMetadata = metaDataMap[field];
        if (fieldMetadata.params.primaryKey) {
            primaryKey = fieldMetadata.params.name;
        }
    }
    var queryObj = query;
    var options = {};
    Enumerable.from(queryObj).forEach((x) => {
        if (exports.filterProps.indexOf(x.key) >= 0) {
            options[x.key] = x.value;
            if (x.key == 'sort') {
                options[x.key] = entityService.getSortCondition(x.value);
            }
            else {
                options[x.key] = x.value;
            }
            delete queryObj[x.key];
        }
        else if (primaryKey && x.key == primaryKey && x.key) {
            let valStr = queryObj[x.key]; //Incoming Comma Separated String Values
            let valArray;
            if (valStr == "" || valStr == null || valStr == "null" || valStr == '""' || valStr == "''" || valStr == undefined || valStr == "undefined") {
                valArray = [];
            }
            else {
                valArray = valStr.split(","); //Convert Comma Separated Values to Array
            }
            queryObj[x.key] = { $in: valArray };
        }
        else {
            var val = queryObj[x.key];
            var i = val.indexOf('%LIKE%');
            if (i == 0) {
                // contains
                val = val.replace('%LIKE%', '');
                queryObj[x.key] = entityService.getLikeCondition(val);
            }
            else {
                i = val.indexOf('%START%');
                if (i == 0) {
                    // starts with
                    val = val.replace('%START%', '');
                    queryObj[x.key] = entityService.getStartsWithCondition(val);
                }
            }
        }
    });
    return { queryObj: queryObj, options: options };
};
exports.filterProps = ['rows', 'start', 'from', 'until', 'order', 'fields', 'skip', 'limit', 'sort', 'lt', 'gt', 'lte', 'gte', 'lt_value', 'gt_value', 'lte_value', 'gte_value'];

//# sourceMappingURL=queryOptions.js.map
